#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "FileSystem.h"



/* Cree un systeme de fichier sur le disque dur
 * 
 * filename	: Nom du fichier
 */
void createfs(const char * filename)
{
	int file = open(filename, O_CREAT | O_EXCL | O_RDWR, S_IWUSR | S_IRUSR);
	if(file == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ouverture de ‘%s’ : ‘%s’.\n", filename, strerror(errno));
		exit(EXIT_FAILURE);
	}
	IndexTable_t *index = (IndexTable_t*) calloc (1,sizeof(IndexTable_t));
	FreeSpaceTable_t *freespacetable = (FreeSpaceTable_t*) calloc (1,sizeof(FreeSpaceTable_t));
	
	if (index==NULL || freespacetable == NULL)
	{
		fprintf(stderr, "Erreur lors de l'allocation des table d'index");
		exit (EXIT_FAILURE);
	}
	
	if(write(file, index, sizeof(IndexTable_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture dans ‘%s’ : ‘%s’.\n", filename, strerror(errno));
		exit(EXIT_FAILURE);
	}
	if(write(file, freespacetable, sizeof(FreeSpaceTable_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture dans ‘%s’ : ‘%s’.\n", filename, strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	if(close(file) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la fermeture du descripteur de fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	
}



/* Ouvre un systeme de fichier du disque dur
 * 
 * fs : Acces au systeme de fichier
 * file	: Systeme de fichier a charger
 */
void openfs(FileSystem_t* fs,const char * file)
{
	// on verifie lerreur de pointeur null
	if(fs == NULL)
	{
		fprintf(stderr, "Le pointeur vers la structure du systeme de fichier est NULL");
		exit(EXIT_FAILURE);
	}
	
	// on ouvre le fichier
	int filedes = open(file, O_RDWR);
	
	if(filedes == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ouverture de ‘%s’ : ‘%s’.\n", file, strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	// on alloue les element du systeme de fichier
	fs->fd = filedes;
	fs->index_table = (IndexTable_t*) calloc (1,sizeof(IndexTable_t));
	fs->free_space_table = (FreeSpaceTable_t*) calloc (1,sizeof(FreeSpaceTable_t));
	
	// on verifie l'allocation
	if (fs->index_table == NULL || fs->free_space_table == NULL)
	{
		fprintf(stderr, "Erreur lors de l'allocation des table d'index");
		exit (EXIT_FAILURE);
	}
	
	// puis on y place l'index de tout les inode en memoire
	if(read(fs->fd, fs->index_table, sizeof(IndexTable_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture dans ‘%s’ : ‘%s’.\n", file, strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	// en en fin on place en memoire les case libre dans le FS
	if(read(fs->fd, fs->free_space_table, sizeof(FreeSpaceTable_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture dans ‘%s’ : ‘%s’.\n", file, strerror(errno));
		exit(EXIT_FAILURE);
	}
	
}

/* Ferme un systeme de fichier 
 * 
 * fs : Acces au systeme de fichier
 */
void closefs(FileSystem_t* fs)
{
	// on desaloue les table d'index correspondant en systeme de fichier que l'on ferme
	free(fs->index_table);
	free(fs->free_space_table);
	
	// on ferme ensuite son descripteur de fichier
	if(close(fs->fd) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la fermeture du descripteur de fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
}

/* Ecrit un fichier sur le systeme de fichier
 * 
 * fs : Acces au systeme de fichier
 * chemin	: chemin du fichier a charger dans le fs
 * chemin_dest_in_fs : nom de destination dans le fs
 */
void writefileinfs(FileSystem_t* fs,const char* chemin, const char * chemin_dest_in_fs)
{
	// on cree une structure pour optenir la taille du fichier
	struct stat fileStat;
	// on recupere effectivement les info du fichier
	stat(chemin,&fileStat);    
	
	off_t block_dispo;
	void *buffer = malloc(BLOCKSIZE);
	off_t offset_first_inode;
	off_t offset_last_adress_table;
	off_t size_fichier = fileStat.st_size;
	int i = 0;
	int taille_a_lire = 0;
	int block = 2;
	
	
	//on recherche si un fichier porte pas deja le meme nom 
	if(search_filename(fs, chemin_dest_in_fs) != -1)
	{
		fprintf(stderr, "Le ficher existe deja dans le system de fichier.\n");
		exit(EXIT_FAILURE);
	}
	
	// on crée notre inode pour le fichier
	inode_t inode_fichier;
	inode_fichier.filename = malloc((strlen(chemin_dest_in_fs)+1)*sizeof(char));
	strcpy(inode_fichier.filename,chemin_dest_in_fs);
	inode_fichier.size = fileStat.st_size;
	inode_fichier.table_adresse = 0;
	
	
	adresse_table_t table_adresse;
	int u;
	for(u = 0; u < NB_ENTRY_DATA_INODE; u++)
	{
		inode_fichier.data[u] = 0;
		table_adresse.data[u] = 0;
	}
	table_adresse.addresse_table_next = 0;
	
	//ouverture du fichier
	int filedes = open(chemin, O_RDONLY);
	if(filedes == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ouverture de ‘%s’ : ‘%s’.\n", chemin, strerror(errno));
		exit(EXIT_FAILURE);
	}

	// Boucle 1 on est dans le data de l'inode !
	do
	{
		//on calcul la taille du bloc qu'il faut
		if((BLOCKSIZE - size_fichier)  <= 0)
		{
			taille_a_lire = BLOCKSIZE;
		}
		else
		{
			taille_a_lire = size_fichier;
		}
		// on lit dans le fichier la taille voulu 
		if(read(filedes, buffer, taille_a_lire) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lorssss de la lecture dans ‘%s’ : ‘%s’.\n", chemin, strerror(errno));
			exit(EXIT_FAILURE);
		}
		
		//on cherche un block libre pour ecrire
		block_dispo = getnextspacefordata(fs,BLOCKSIZE);
		setdataremoved(fs,block_dispo,BLOCKSIZE);
		
		// on place le curseur au bon endroit
		if(lseek(fs->fd, block_dispo, SEEK_SET) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lorsttt de le lecture du fichier : ‘%s’.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		// on ecrit le block
		if(write(fs->fd, buffer, BLOCKSIZE) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
			exit(EXIT_FAILURE);
		}
		inode_fichier.data[i] = block_dispo;
		
		// en enleve le block ecrit de la taille total du fichier
		size_fichier -= taille_a_lire;
		i++;
		
	}while (i < NB_ENTRY_DATA_INODE &&  size_fichier > 0);
	
	//on ecrit l'inode de fichier
	offset_first_inode = getnextspacefordata(fs,sizeof(inode_t)+sizeof(int)+ strlen(inode_fichier.filename)+1);
	setdataremoved(fs,offset_first_inode,sizeof(inode_t)+sizeof(int)+ strlen(inode_fichier.filename)+1);
	write_inode_on_fs(fs->fd,offset_first_inode,&inode_fichier);
	
	i = 0;
	
	// Boucle 2 on est dans une table d'adresse le fichier est trop gros pour tenir que dans l'inode !
	while( size_fichier > 0)
	{
		//on calcul la taille du bloc qu'il faut
		if((BLOCKSIZE - size_fichier)  <= 0)
		{
			taille_a_lire = BLOCKSIZE;
		}
		else
		{
			taille_a_lire = size_fichier;
		}
		// on lit dans le fichier la taille voulu 
		if(read(filedes, buffer, taille_a_lire) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture dans ‘%s’ : ‘%s’.\n", chemin, strerror(errno));
			exit(EXIT_FAILURE);
		}
		
		//on cherche un block libre pour ecrire
		block_dispo = getnextspacefordata(fs,BLOCKSIZE);
		setdataremoved(fs,block_dispo,BLOCKSIZE);
		
		// on place le curseur au bon endroit
		if(lseek(fs->fd, block_dispo, SEEK_SET) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lorspppp de le lecture du fichier : ‘%s’.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		// on ecrit le block
		if(write(fs->fd, buffer, BLOCKSIZE) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
			exit(EXIT_FAILURE);
		}
		table_adresse.data[i] = block_dispo;
		size_fichier -= taille_a_lire;
		
		// si on arrive a la fin du fichier ou que ta table d'adresse a plus de place
		if(i == (NB_ENTRY_DATA_INODE-1) || size_fichier == 0)
		{
			
			
			// on recupere un espace disponible
			off_t espace_dispo = getnextspacefordata(fs,sizeof(adresse_table_t));
			
			// on supprime cette espace espace libre de la table
			setdataremoved(fs,espace_dispo,sizeof(adresse_table_t));
			
			// on ecrit l'inode sur le disque
			write_adresse_table_on_fs(fs->fd,espace_dispo,&table_adresse);

			// on a ecrit le premier block d'adressage on update donc l'anciene table pour pointer vers celle nouvelement ecrite
			if(block == 2)
			{
				inode_fichier.table_adresse = espace_dispo;
				write_inode_on_fs(fs->fd,offset_first_inode,&inode_fichier);
				offset_last_adress_table = espace_dispo;
			}
			else
			{
				adresse_table_t var = read_adresse_table_on_fs(fs->fd,offset_last_adress_table);
				var.addresse_table_next = espace_dispo;
				write_adresse_table_on_fs(fs->fd,offset_last_adress_table,&var);	    
				offset_last_adress_table = espace_dispo;
				
			}
			
			// on a ecrit la table d'adresse maintenant on reinitialise tout a 0
			for(u = 0; u < NB_ENTRY_DATA_INODE; u++)
			{
			  table_adresse.data[u] = 0;
			}
			table_adresse.addresse_table_next = 0;
			
			block++;
		}
		// on remet a 0 si la table d'adresse est pleine sinon on incremente
		if(i == (NB_ENTRY_DATA_INODE-1))
		{
			i = 0;
		}
		else
		{
			i++;
		}
		
	}
	addinodeonindex(fs,offset_first_inode);
	free(buffer);
	free(inode_fichier.filename);
}

/* Supprime un fichier dans le systeme de fichier
 * 
 * fs : Acces au systeme de fichier
 * chemin	: nom du fichier a supprimer
 */
void removefileinfs(FileSystem_t*fs ,const char* chemin)
{
	// on recherche si le fichier existe
	off_t emplacement_inode = search_filename(fs, chemin);
	if(emplacement_inode == -1)
	{
		fprintf(stderr, "Le fichier n'a pas été trouver dans le systeme de fichier.\n");
		exit(EXIT_FAILURE);
	}
	
	off_t emplacement_table_adresse = 0;
	inode_t inode_fichier = read_inode_on_fs(fs->fd,emplacement_inode);
	adresse_table_t table;
	int i = 0;

	// ajoute l'inode a la table de liberation 
	insertdataremoved(fs,emplacement_inode,sizeof(inode_t)+sizeof(unsigned int)+((strlen(inode_fichier.filename)+1)*sizeof(char)));
	// on dereference l'inode sur le disque 
	deleteinodeonindex(fs->fd,fs->index_table,emplacement_inode);
	
	off_t size_fichier = inode_fichier.size;
	//une premiere boucle pour les data de linode ( pour petit fichier)
	do
	{
		
		insertdataremoved(fs,inode_fichier.data[i],BLOCKSIZE);
		size_fichier -= BLOCKSIZE;
		i++;
		
	}while (i < NB_ENTRY_DATA_INODE &&  size_fichier > 0);
	

	emplacement_table_adresse = inode_fichier.table_adresse;
	i =0;
	
	//si on rentre dans cette boucle cest quon a un gros fichier
	while( size_fichier > 0)
	{
		if(i ==0 )
		{
			insertdataremoved(fs,emplacement_table_adresse,sizeof(adresse_table_t));
			table = read_adresse_table_on_fs(fs->fd,emplacement_table_adresse);
			emplacement_table_adresse = table.addresse_table_next;  
		}
		// pour chaque table on ajoute chaque block de donné a la table de liberation
		insertdataremoved(fs,table.data[i],BLOCKSIZE);
		size_fichier -= BLOCKSIZE;
		
		if(i == (NB_ENTRY_DATA_INODE-1))
		{
			i = 0;
		}
		else
		{
			i++;
		}
		
	}
	free(inode_fichier.filename);
}

/* Ecrit un fichier du systeme de fichier vers le disque dur
 * 
 * fs : Acces au systeme de fichier
 * filename	: nom du fichier a extraire
 * chemin_sortit : nom du fichier en destionation
 */
void writefiletosystem(FileSystem_t *fs ,const char* filename,const char * chemin_sortit)
{
	
	void *buffer = malloc(BLOCKSIZE);
	
	// on recherche si le fichier existe
	off_t emplacement_inode = search_filename(fs, filename);
	if(emplacement_inode == -1)
	{
		fprintf(stderr, "Le fichier n'a pas été trouver dans le systeme de fichier.\n");
		exit(EXIT_FAILURE);
	}
	
	off_t emplacement_table_adresse = 0;
	inode_t inode_fichier = read_inode_on_fs(fs->fd,emplacement_inode);
	adresse_table_t table;
	int i = 0;

	

	//ouverture du fichier
	int filedest = open(chemin_sortit, O_CREAT | O_EXCL | O_RDWR,S_IWUSR | S_IRUSR);
	if(filedest == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ouverture de ‘%s’ : ‘%s’.\n", chemin_sortit, strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	off_t size_fichier = inode_fichier.size;
	long taille_a_lire = 0;
	//une premiere boucle pour les data de linode ( pour petit fichier)
	do
	{
		if((BLOCKSIZE - size_fichier)  <= 0)
		{
			taille_a_lire = BLOCKSIZE;
		}
		else
		{
			taille_a_lire = size_fichier;
		}     
		if(lseek(fs->fd, inode_fichier.data[i], SEEK_SET) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if(read(fs->fd, buffer, taille_a_lire) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture : ‘%s’.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if(write(filedest, buffer, taille_a_lire) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
			exit(EXIT_FAILURE);
		}
		size_fichier -= taille_a_lire;
		i++;
		
	}while (i <NB_ENTRY_DATA_INODE &&  size_fichier > 0);
	
	emplacement_table_adresse = inode_fichier.table_adresse;
	i =0;
	
	//si on rentre dans cette boucle cest quon a un gros fichier
	while( size_fichier > 0)
	{
		if(i ==0 )
		{
			table = read_adresse_table_on_fs(fs->fd,emplacement_table_adresse);
			emplacement_table_adresse = table.addresse_table_next;
			
		}
		if((BLOCKSIZE - size_fichier)  <= 0)
		{
			taille_a_lire = BLOCKSIZE;
		}
		else
		{
			taille_a_lire = size_fichier;
		}	
		if(lseek(fs->fd, table.data[i], SEEK_SET) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if(read(fs->fd, buffer, taille_a_lire) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if(write(filedest, buffer, taille_a_lire) == -1)
		{
			fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
			exit(EXIT_FAILURE);
		}
		size_fichier -= taille_a_lire;
		if(i == (NB_ENTRY_DATA_INODE-1))
		{
			i = 0;
		}
		else
		{
			i++;
		}
		
	}
	free(buffer);
	free(inode_fichier.filename);
}

/* Affiche tout les fichier du system de fichier
 * 
 * fs : Acces au systeme de fichier
 */
void ListAllFile(FileSystem_t *fs)
{
	IndexTable_t table = *(fs->index_table);
	inode_t inode;
	int firsttour = 1;
	int i;
	do
	{
		if(firsttour)
		{
			firsttour = 0;
		}
		else
		{
			table = getindextable(fs->fd,table.indexNext);
		}

		for(i = 0; i < NB_ENTRY_INDEX_TABLE; i++)
		{
			if(table.inode[i] != 0)
			{
				inode = read_inode_on_fs(fs->fd,table.inode[i]);
				printf("%s\n",inode.filename);
				free(inode.filename);
			}
		}
	}while(table.indexNext != 0);

}

/* Defragmente tout les object du systeme de fichier
 * 
 * fs : Acces au systeme de fichier
 */
void defragment(FileSystem_t *fs)
{
  int lenghtofdata = 0;
  off_t parent = 0; // sert a avoir l'adresse de l'inode pour retrouver l'adresse des data originel
  void *buffer;
  enum type type_search = inode;
  
  
  
  /****************************************** DEFRAGMENT INODE ***********************************************/
  
  // on recherche le plus lointin inode 
  off_t last_data = search_last_object_in_fs( fs,type_search,&parent);
  
  // si on en a pas on risque un segfault on va pas plus loin
  if(last_data != -1)
  {
    // sinon on cherche la taille de l'inode
    lenghtofdata = sizeofinode(fs->fd,last_data);
  }
  // on regarde les espace les plus pret
  off_t first_space = getnearspacefordata(fs, lenghtofdata);
  
  // t'en que on a un inode qui est plus loin que le plus pret des espace
  while (last_data > first_space && first_space != -1)
  {
    printf("Inode defragmented : %ld  to dest : %ld\n",(long)last_data,(long)first_space);
    //on lit l'inode
    inode_t inode = read_inode_on_fs(fs->fd,last_data);
    //on le reecrit dans l'espace libre trouver
    write_inode_on_fs(fs->fd,first_space,&inode);
    // on suprime l'espace libre de la table despace libre
    setdataremoved(fs,first_space,lenghtofdata);
    //on edit l'index pour le changement d'emplacement de l'inode
    editInodeIndexTable(fs,last_data,first_space); 
    // et on libre la memoire
    free(inode.filename);
    
    // on ajoute l'espace liberer a l'espace libre
    insertdataremoved(fs,last_data,lenghtofdata);
    
    // et on recommence la recherche
    last_data = search_last_object_in_fs( fs,type_search,&parent);
    lenghtofdata = sizeofinode(fs->fd,last_data);
    first_space = getnearspacefordata(fs, lenghtofdata);
  }
  
  
  /* ********************************************DEFRAGMENT DATA *********************************************************/
  
  type_search = data;
  
  last_data = search_last_object_in_fs(fs, type_search ,&parent);
  lenghtofdata = BLOCKSIZE;
  first_space = getnearspacefordata(fs, lenghtofdata);
  
  buffer = malloc(BLOCKSIZE);

  
  while (last_data > first_space && first_space != -1)
  {
    printf("Data defragmented : %ld  to dest : %ld\n",(long)last_data,(long)first_space);
    editoffsetofdata(fs->fd,parent,last_data,first_space);
    
    if(lseek(fs->fd, last_data, SEEK_SET) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(read(fs->fd, buffer, lenghtofdata) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(lseek(fs->fd, first_space, SEEK_SET) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(write(fs->fd, buffer, lenghtofdata) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
	    exit(EXIT_FAILURE);
    }

    // on suprime l'espace libre de la table despace libre
    setdataremoved(fs,first_space,lenghtofdata);
    
    // on ajoute l'espace liberer a l'espace libre
    insertdataremoved(fs,last_data,lenghtofdata);
    
    // et on recommence la recherche
    last_data = search_last_object_in_fs(fs,type_search,&parent);
    first_space = getnearspacefordata(fs, lenghtofdata);
  }
  free(buffer);
  
  /* ***********************************DEFRAGMENT TABLE ADRESSE ***************************************************/
  
  type_search = adressetable;
  last_data = search_last_object_in_fs(fs, type_search ,&parent);
  lenghtofdata = sizeof(adresse_table_t);
  first_space = getnearspacefordata(fs, lenghtofdata);
  buffer = malloc(lenghtofdata);
  
  while (last_data > first_space && first_space != -1)
  {
    printf("Adresse table defragmented : %ld  to dest : %ld\n",(long)last_data,(long)first_space);
    editoffsetofadressetable(fs->fd,parent,last_data,first_space);
    
    if(lseek(fs->fd, last_data, SEEK_SET) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(read(fs->fd, buffer, lenghtofdata) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(lseek(fs->fd, first_space, SEEK_SET) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(write(fs->fd, buffer, lenghtofdata) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
	    exit(EXIT_FAILURE);
    }
    
    // on suprime l'espace libre de la table despace libre
    setdataremoved(fs,first_space,lenghtofdata);
    
    // on ajoute l'espace liberer a l'espace libre
    insertdataremoved(fs,last_data,lenghtofdata);
    
    // et on recommence la recherche
    last_data = search_last_object_in_fs(fs,type_search,&parent);
    first_space = getnearspacefordata(fs, lenghtofdata);
  }
  free(buffer);
  
  
  /* ***********************************DEFRAGMENT INDEX TABLE ***************************************************/
  
  type_search = indextable;
  last_data = search_last_object_in_fs(fs, type_search ,&parent);
  lenghtofdata = sizeof(IndexTable_t);
  first_space = getnearspacefordata(fs, lenghtofdata);
  buffer = malloc(lenghtofdata);
  while (last_data > first_space && first_space != -1)
  {
    printf("Table d'index defragmented : %ld  to dest : %ld\n",(long)last_data,(long)first_space);
    editIndexTable(fs,last_data,first_space);
	
    if(lseek(fs->fd, last_data, SEEK_SET) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(read(fs->fd, buffer, lenghtofdata) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(lseek(fs->fd, first_space, SEEK_SET) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(write(fs->fd, buffer, lenghtofdata) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
	    exit(EXIT_FAILURE);
    }
    // on suprime l'espace libre de la table despace libre
    setdataremoved(fs,first_space,lenghtofdata);
    
    // on ajoute l'espace liberer a l'espace libre
    insertdataremoved(fs,last_data,lenghtofdata);
    
    // et on recommence la recherche
    last_data = search_last_object_in_fs(fs,type_search,&parent);
    first_space = getnearspacefordata(fs, lenghtofdata);
    
  }
  free(buffer);
  
  
  
  /* ***********************************DEFRAGMENT FREE SPACETABLE ***************************************************/
  
  last_data = search_last_freespacetable_in_fs(fs);
  lenghtofdata = sizeof(FreeSpaceTable_t);
  first_space = getnearspacefordata(fs, lenghtofdata);
  buffer = malloc(lenghtofdata);
  while (last_data > first_space && first_space != -1)
  {
    printf("Space Table defragmented : %ld  to dest : %ld\n",(long)last_data,(long)first_space);
    editFreeSpaceTable(fs,last_data,first_space);
	
    if(lseek(fs->fd, last_data, SEEK_SET) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(read(fs->fd, buffer, lenghtofdata) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(lseek(fs->fd, first_space, SEEK_SET) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
	    exit(EXIT_FAILURE);
    }
    if(write(fs->fd, buffer, lenghtofdata) == -1)
    {
	    fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
	    exit(EXIT_FAILURE);
    }
    // on suprime l'espace libre de la table despace libre
    setdataremoved(fs,first_space,lenghtofdata);
    
    // on ajoute l'espace liberer a l'espace libre
    insertdataremoved(fs,last_data,lenghtofdata);
    
    // et on recommence la recherche
    last_data = search_last_freespacetable_in_fs(fs);
    first_space = getnearspacefordata(fs, lenghtofdata);
    
  }
  free(buffer);
  
  optimise_file(fs);

}

/* Suprimme toute la fin du fichier inutilisé. 
 * Il recherche le dernier object dans le fichier.
 * Puis va suprimmer tout l'espace libre non utilisé après.
 * et enfin va suprimmer l'espace libre de la table d'espace libre
 * 
 * fs : Acces au systeme de fichier
 */
void optimise_file(FileSystem_t *fs)
{
  off_t last_object = 0;
  int type = -1;  // 0 FreeSpaceTable_t, 1 data, 2 adresse table,3 inode, 4 indextable
  
  
  
  
  off_t last_data = search_last_freespacetable_in_fs(fs);
  if(last_data > last_object)
  {
    last_object = last_data;
    type = 0;
  }
  
  enum type type_search = inode;
  last_data = search_last_object_in_fs(fs, type_search ,NULL);
  if(last_data > last_object)
  {
    last_object = last_data;
    type = 3;
  }
  
  type_search = data;
  last_data = search_last_object_in_fs(fs, type_search ,NULL);
  if(last_data > last_object)
  {
    last_object = last_data;
    type = 1;
  }
  
  
  type_search = adressetable;
  last_data = search_last_object_in_fs(fs, type_search ,NULL);
  if(last_data > last_object)
  {
    last_object = last_data;
    type = 2;
  }
  
  type_search = indextable;
  last_data = search_last_object_in_fs(fs, type_search ,NULL);
  if(last_data > last_object)
  {
    last_object = last_data;
    type = 4;
  }
  
  int length = 0;
  if(type == 0){length = sizeof(FreeSpaceTable_t);}
  else if(type == 3){length = sizeofinode(fs->fd,last_object);}
  else if(type == 2){length = sizeof(adresse_table_t);}
  else if(type == 4){length = sizeof(IndexTable_t);}
  else if(type == 1){length = BLOCKSIZE;}
  
  ftruncate(fs->fd, last_object+length);
  
  //puis on suprimme tout l'espace libre qui depase cette valeur
  removeallfreespaceafter(fs,last_object+length);
  
  
}
