#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "FreeSpaceTable.h"

/* Renvoie un offset libre pouvant acceuillir les données
 * 
 * fs	: permet d'acceder au system de fichier
 * lenghtofdata	: taille minimal de la case pour les donner 
 */
off_t getnextspacefordata(FileSystem_t * fs, int lenghtofdata) 
{
	// on charge la table d'espace libre en memoire
	FreeSpaceTable_t table = *(fs->free_space_table);
	int i = 0;
	int firsttour = 1;
	
	//on parcour toute la table d'espace libre
	do
	{
		// sauf pour le premier tour a chaque tour on charge une nouvelle table d'espace libre
		if(firsttour) 
		{
			firsttour = 0;
		}
		else
		{
			table = loadfreespacetable(fs->fd,table.next_free_space_table);
		}
		
		for(i = 0; i < NB_ENTRY_FREE_SPACE_TABLE; i++)
		{
		    // si on a un espace libre de dispo, on l'utilise
			if(table.lenght[i] >= lenghtofdata)
			{
				return table.freespace[i]; 
			}
		}
		
	}while(table.next_free_space_table != 0);

	// si on ne trouve pas d'espace libre on considere que c'est la fin du fichier ou on doit ecrire
	off_t var = lseek(fs->fd, 0, SEEK_END);
	if(var == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	return var;

}
/*Renvoie l'offset libre le plus proche pouvant acceuillir les donner
 * 
 * fs	: permet d'acceder au system de fichier
 * lenghtofdata	: taille minimal de la case pour les donner 
 */
off_t getnearspacefordata(FileSystem_t * fs, int lenghtofdata) 
{
	// on charge la table d'espace libre en memoire
	FreeSpaceTable_t temp = *(fs->free_space_table);
	int i = 0;
	int firsttour = 1;
	off_t espace_le_plus_proche = 0;
	
	//on parcour toute la table d'espace libre
	do
	{
		if(firsttour)
		{
			firsttour = 0;
		}
		else
		{
			temp = loadfreespacetable(fs->fd,temp.next_free_space_table);
		}
		
		for(i = 0; i < NB_ENTRY_FREE_SPACE_TABLE; i++)
		{
		    // si on trouve une case libre plus proche de l'anciene on redefinie l'espace le plus proche
			if(temp.lenght[i] >= lenghtofdata)
			{
				if(espace_le_plus_proche > temp.freespace[i] || espace_le_plus_proche == 0)
				{
					espace_le_plus_proche = temp.freespace[i];
				}
			}
		}
		
	}while(temp.next_free_space_table != 0);
	
	// si l'espace le plus proche na pas plus etre trouver c'est qu'on a plus de place nul par on renvoie -1
	if(espace_le_plus_proche == 0)
	{
		return -1;
	}
	return espace_le_plus_proche;
}
/* Ordonne a la table d'espace libre de suprimer l'espace libre définie a l'offset et de longueur spécifier
 * Attention : si il reste des donner libre apres la longueur lenght, cela modifie la table d'espace libre pour lui ajouter ces block 
 * 
 * fs	: Permet d'acceder au system de fichier
 * offset	: Endroit ou l'espace libre commence
 * lenght	: Longueur des donner que l'on va y placer 
 */
void setdataremoved(FileSystem_t * fs,off_t offset,int lenght)
{
	FreeSpaceTable_t temp = *(fs->free_space_table);
	int i = 0;
	int firsttour = 1;
	off_t offsetcurindextable=sizeof(IndexTable_t); // VERIFIERRRRRR
	do
	{
		if(firsttour == 0)
		{
			offsetcurindextable = temp.next_free_space_table;
			temp = loadfreespacetable(fs->fd,temp.next_free_space_table);
		}
		
		for(i = 0; i < NB_ENTRY_FREE_SPACE_TABLE; i++)
		{
			// quand on trouve loffset que l'on veut suprimer alors on le met a 0 et on ecrit la modif
			if(temp.freespace[i] == offset)
			{
				int var = temp.lenght[i] - lenght;
				if(var != 0)
				{ // il reste de lespace a utilisé
					temp.freespace[i] += lenght;
					temp.lenght[i] -= lenght; 
				}
				else
				{
					temp.freespace[i] = 0;
					temp.lenght[i] = 0; 
				}
				if(firsttour)
				{
					*(fs->free_space_table) = temp;
				}
				writeSpaceTable(fs->fd,offsetcurindextable,temp);
				return ;
			}
		}
		if(firsttour)
		{
			firsttour = 0;
		}
		
	}while(temp.next_free_space_table != 0);

}
/* Ordonne a la table d'espace libre d'ajouter de l'espace libre définie a l'offset et de longueur spécifier 
 * 
 * fs	: Permet d'acceder au system de fichier
 * offset	: Endroit ou l'espace libérer va etre
 * lenght	: Longueur de l'espace libre 
 */
void insertdataremoved(FileSystem_t * fs,off_t offset,int lenght)
{
	off_t offsetcurindextable=sizeof(IndexTable_t);
	FreeSpaceTable_t temp = *(fs->free_space_table);
	int i = 0;
	int firsttour = 1;
	do
	{
		if(firsttour == 0) 
		{
			offsetcurindextable = temp.next_free_space_table;
			temp = loadfreespacetable(fs->fd,temp.next_free_space_table);
		}
		
		for(i = 0; i < NB_ENTRY_FREE_SPACE_TABLE; i++)
		{
			// quand on trouve loffset que l'on veut suprimer alors on le met a 0 et on ecrit la modif
			if(temp.freespace[i] == 0)
			{
				temp.freespace[i] = offset;
				temp.lenght[i] = lenght;
				// on update la table despace en memoire si on est au premier tour
				if(firsttour)
				{
					*(fs->free_space_table) = temp;
				}
				writeSpaceTable(fs->fd,offsetcurindextable,temp);
				return ;
			}
		}
		if(firsttour)
		{
			firsttour = 0;
		}
		
	}while(temp.next_free_space_table != 0);
	
	
	temp.next_free_space_table = getnextspacefordata(fs, sizeof(FreeSpaceTable_t));
	setdataremoved(fs,temp.next_free_space_table,sizeof(FreeSpaceTable_t));
	
	writeSpaceTable(fs->fd,offsetcurindextable,temp);
	
	if(offsetcurindextable == sizeof(IndexTable_t))
	{
		*(fs->free_space_table) = temp;
	}
	
	FreeSpaceTable_t next;
	for(i = 0; i < NB_ENTRY_FREE_SPACE_TABLE; i++)
	{
		next.freespace[i] = 0;
		next.lenght[i] = 0;
	}
	next.next_free_space_table = 0;
	
	next.freespace[0] = offset;
	next.lenght[0] = lenght;
	writeSpaceTable(fs->fd,temp.next_free_space_table,next);
	

}
/* Ecrit une table d'espace libre sur le disque virtuelle
 * 
 * fs	: Permet d'acceder au system de fichier
 * offset	: Endroit ou ecrire la table
 */
void writeSpaceTable(int fd,off_t offset,FreeSpaceTable_t table)
{
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	// puis on ecrit notre table d'index
	if(write(fd, &table, sizeof(FreeSpaceTable_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",  strerror(errno));
		exit(EXIT_FAILURE);
	}
	
}
/* Renvoie la table d'espace libre de l'offset spécifier
 * 
 * fs	: Permet d'acceder au system de fichier
 * offset	: Endroit ou lire la table
 */
FreeSpaceTable_t loadfreespacetable(int fd,off_t offset)
{
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	// on alloue la memoire pour les donnée
	FreeSpaceTable_t  table;
	
	// en en fin on place en memoire les case libre dans le FS
	if(read(fd, &table, sizeof(FreeSpaceTable_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	return table;
}
/* Affiche la table d'espace libre
 * 
 * table	: table d'espace libre a afficher
 */
void showFreeSpaceTable(FreeSpaceTable_t * table)
{
	int i  = 0;
	printf("Taille d'une table d'espace libre : %ld\n", sizeof(FreeSpaceTable_t));
	for(i = 0; i < NB_ENTRY_FREE_SPACE_TABLE; i++)
	{
		printf("Espace libre %d : %ld longueur : %d\n",i,(long)table->freespace[i],table->lenght[i]);
	}
	printf("Table suivante %ld \n",(long)table->next_free_space_table);

}
/* Renvoie la table d'espace libre la plus loin dans le systeme de fichier
 * 
 * fs	: Permet d'acceder au system de fichier
 */
off_t search_last_freespacetable_in_fs(FileSystem_t * fs)
{
	FreeSpaceTable_t table = *(fs->free_space_table);
	
	int firsttour = 1;
	off_t last = 0;
	do
	{
		if(firsttour)
		{
			firsttour = 0;
		}
		else
		{
			table = loadfreespacetable(fs->fd,table.next_free_space_table);
		}
		if(table.next_free_space_table > last)
		{
		    last = table.next_free_space_table;
		}
		
		
	}while(table.next_free_space_table != 0);
	
	if(last == 0)
	{
	  return -1;
	}
	return last;
  
}
/* Modifier dans la chaine de table d'espace libre, 
 * redéfinie les table pour que la table etant a l'ancien offset puissent pointer vers le nouvelle ofset de cette table
 * 
 * fs	: Permet d'acceder au system de fichier
 * ancien_offset : ancienne offset d'une table d'espace libre
 * new_offset : nouvelle offset de cette table
 */
void editFreeSpaceTable(FileSystem_t * fs,off_t ancien_offset,off_t new_offset)
{
	FreeSpaceTable_t table = *(fs->free_space_table);
	
	int firsttour = 1;
	off_t offsetcurtable = 0;
	do
	{
		
		if(firsttour == 0)
		{
			offsetcurtable = table.next_free_space_table;
			table = loadfreespacetable(fs->fd,table.next_free_space_table);
		}
		
		if(firsttour && table.next_free_space_table == ancien_offset )
		{
		  table.next_free_space_table = new_offset;
		  *(fs->free_space_table) = table;
		  writeSpaceTable(fs->fd,sizeof(IndexTable_t),table);

		  return ;
		}
		else if (table.next_free_space_table == ancien_offset)
		{
		  table.next_free_space_table = new_offset;
		  writeSpaceTable(fs->fd,offsetcurtable,table);
		  return ;
		}
		
		if(firsttour)
		{
			firsttour = 0;
		}
		
	}while(table.next_free_space_table != 0);
	
  }
/* On suprimme tout les espace libre apres offset 
 * 
 * fs	: Permet d'acceder au system de fichier
 * offset : emplacement 
 */
  void removeallfreespaceafter(FileSystem_t * fs,off_t offset)
  {
	FreeSpaceTable_t temp = *(fs->free_space_table);
	int i = 0;
	int firsttour = 1;
	off_t offsetcurindextable=sizeof(IndexTable_t); // VERIFIERRRRRR
	do
	{
		if(firsttour == 0)
		{
			offsetcurindextable = temp.next_free_space_table;
			temp = loadfreespacetable(fs->fd,temp.next_free_space_table);
		}
		
		for(i = 0; i < NB_ENTRY_FREE_SPACE_TABLE; i++)
		{
			// quand on trouve loffset que l'on veut suprimer alors on le met a 0 et on ecrit la modif
			if(temp.freespace[i] >= offset)
			{
				temp.freespace[i] = 0;
				temp.lenght[i] = 0; 
				if(firsttour)
				{
					*(fs->free_space_table) = temp;
				}
				writeSpaceTable(fs->fd,offsetcurindextable,temp);
			}
		}
		if(firsttour)
		{
			firsttour = 0;
		}
		
	}while(temp.next_free_space_table != 0);
  }