#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "FileSystem.h"

/* Ajoute un inode dans la table d'index
 * cette indode est placer au premier espace libre ou dans une table d'index nouvellement cree si il n'y a plus de place
 * 
 * fs	: Permet d'acceder au system de fichier
 * offset	: adresse de l'indode dans le fichier
 */
void addinodeonindex(FileSystem_t * fs,off_t offset)
{
	
	IndexTable_t temp = *(fs->index_table);
	int i = 0;
	int firsttour = 1;
	off_t offsetcurindextable=0;
	do
	{
		
		if(firsttour == 0)
		{
			offsetcurindextable = temp.indexNext;
			temp = getindextable(fs->fd,temp.indexNext);
		}
		
		for(i = 0; i < NB_ENTRY_INDEX_TABLE; i++)
		{
			if(temp.inode[i] == 0)
			{
				temp.inode[i] = offset;
				if(firsttour)
				{
					*(fs->index_table) = temp;
				}
				writeIndexTable(fs->fd,offsetcurindextable,temp);
				return ;
			}
		}
		if(firsttour)
		{
			firsttour = 0;
		}
		
	}while(temp.indexNext != 0);
	
	
	// recupere un espace disponible
	off_t offset_to_write = getnextspacefordata(fs,sizeof(IndexTable_t));  
	
	// on fait pointer le dernier index vers le prochain
	temp.indexNext = offset_to_write;
	
	// on ecrit la modification
	writeIndexTable(fs->fd,offsetcurindextable,temp);
	
	// on remet a 0 l'index table temporaire
	memset ( &temp, 0, sizeof(IndexTable_t) );
	
	// on ajoute a la 
	temp.inode[0] = offset;
	
	// on ecrit le prochain index ou il doit etre ecrit
	writeIndexTable(fs->fd,offset_to_write,temp);
	
	// on met a jour la table d'espace libre
	setdataremoved(fs,offset_to_write,sizeof(IndexTable_t));
	
	
}
/* Suprimme un inode dans la table d'index
 * 
 * fd	: Descripteur de fichier du systeme de fichier
 * table : table ou l'on commencera les recherche du fichier a suprimmer 
 * offset	: adresse de l'indode du fichier a suprimmer
 */
void deleteinodeonindex(int fd,IndexTable_t *table,off_t offset)
{
	IndexTable_t temp = *(table);
	int i = 0;
	int firsttour = 1;
	off_t offsetcurindextable=0;
	do
	{
		if(firsttour)
		{
			firsttour = 0;
		}
		else
		{
			offsetcurindextable = temp.indexNext;
			temp = getindextable(fd,temp.indexNext);
		}
		
		for(i = 0; i < NB_ENTRY_INDEX_TABLE; i++)
		{
			// quand on trouve loffset que l'on veut suprimer alors on le met a 0 et on ecrit la modif
			if(temp.inode[i] == offset)
			{
				temp.inode[i] = 0;
				writeIndexTable(fd,offsetcurindextable,temp);
				return ;
			}
		}
		
	}while(temp.indexNext != 0);	
}
/* Ecrit une table d'index dans le fichier a l'offset donné
 * 
 * fd	: Descripteur de fichier du systeme de fichier
 * offset	: adresse ou ecrire la table
 * table : table a ecrire
 */
void writeIndexTable(int fd,off_t offset,IndexTable_t table)
{
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	// puis on ecrit notre table d'index
	if(write(fd, &table, sizeof(IndexTable_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
		exit(EXIT_FAILURE);
	}
	
}
/* lit et renvoie une table d'index dans le fichier a l'offset donné
 * 
 * fd	: Descripteur de fichier du systeme de fichier
 * offset	: adresse ou lire la table
 */
IndexTable_t getindextable(int fd,off_t offset)
{ 
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	// on alloue la memoire pour les donnée
	IndexTable_t  table;
	
	// en en fin on place en memoire les case libre dans le FS
	if(read(fd, &table, sizeof(IndexTable_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	return table;
}
/* affiche une table d'index passer en parametre
 * 
 * table	: Table d'index a afficher
 */
void showIndexTable(IndexTable_t * table)
{
	int i  = 0;
	printf("Taille d'un Index : %ld\n", sizeof(IndexTable_t));
	for(i = 0; i < NB_ENTRY_INDEX_TABLE; i++)
	{
		printf("Inode %d : %ld\n",i,(long)table->inode[i]);
	}
	printf("Inode next %ld \n",(long)table->indexNext);

}
/* affiche toute les table d'index du systeme de fichier
 * 
 * fs	: Acces au systeme de fichier
 */
void showallIndexTable(FileSystem_t * fs)
{
	IndexTable_t table = *(fs->index_table);
	int firsttour = 1;
	int i  = 0;
	do
	{
		if(firsttour)
		{
			firsttour = 0;
		}
		else
		{
			table = getindextable(fs->fd,table.indexNext);
		}
		
		for(i = 0; i < NB_ENTRY_INDEX_TABLE; i++)
		{
			printf("Inode %d : %ld\n",i,(long)table.inode[i]);
		}
		printf("Inode next %ld \n",(long)table.indexNext);
	}while(table.indexNext != 0);

}
/* Modifier l'adresse de l'inode dans la table d'index
 * 
 * fs	: Acces au systeme de fichier
 * ancien_offset : Ancienne adresse de l'inode
 * new_offset : Nouvelle adresse de l'inode
 */
void editInodeIndexTable(FileSystem_t * fs,off_t ancien_offset,off_t new_offset)
{
	IndexTable_t temp = *(fs->index_table);
	int i = 0;
	int firsttour = 1;
	off_t offsetcurindextable=0;
	do
	{
		
		if(firsttour == 0)
		{
			offsetcurindextable = temp.indexNext;
			temp = getindextable(fs->fd,temp.indexNext);
		}
		
		for(i = 0; i < NB_ENTRY_INDEX_TABLE; i++)
		{
			if(temp.inode[i] == ancien_offset)
			{
				temp.inode[i] = new_offset;
				if(firsttour)
				{
					*(fs->index_table) = temp;
				}
				writeIndexTable(fs->fd,offsetcurindextable,temp);
				return ;
			}
		}
		if(firsttour)
		{
			firsttour = 0;
		}
		
	}while(temp.indexNext != 0);
  
  
}
/* Modifier l'adresse d'une table d'index
 * 
 * fs	: Acces au systeme de fichier
 * ancien_offset : Ancienne adresse de la table d'index
 * new_offset : Nouvelle adresse de la table d'index
 */
void editIndexTable(FileSystem_t * fs,off_t ancien_offset,off_t new_offset)
{
  IndexTable_t temp = *(fs->index_table);
  int firsttour = 1;
  off_t offsetcurindextable=0;
  do
  {
	  
	  if(firsttour == 0)
	  {
		  offsetcurindextable = temp.indexNext;
		  temp = getindextable(fs->fd,temp.indexNext);
	  }
	  
	  
	  if(firsttour && temp.indexNext == ancien_offset )
	  {
	    temp.indexNext = new_offset;
	    *(fs->index_table) = temp;
	    writeIndexTable(fs->fd,offsetcurindextable,temp);
	    return ;
	  }
	  else if (temp.indexNext == ancien_offset)
	  {
	    temp.indexNext = new_offset;
	    writeIndexTable(fs->fd,offsetcurindextable,temp);
	    return ;
	  }
	  

	  if(firsttour)
	  {
		  firsttour = 0;
	  }
	  
  }while(temp.indexNext != 0);
  
  
}
/* Renvoie l'offset du nom de fichier passé en prarametre -1 si introuvable
 * 
 * fs	: Acces au systeme de fichier
 * filename : nom du fichier
 */
off_t search_filename(FileSystem_t * fs, const char *filename)
{
	IndexTable_t table = *(fs->index_table);
	inode_t inode;
	int firsttour = 1;
	int i;
	do
	{
		if(firsttour)
		{
			firsttour = 0;
		}
		else
		{
			table = getindextable(fs->fd,table.indexNext);
		}
		
		for(i = 0; i < NB_ENTRY_INDEX_TABLE; i++)
		{
			if(table.inode[i] != 0)
			{
				inode = read_inode_on_fs(fs->fd,table.inode[i]);
				if(strcmp(filename,inode.filename) == 0)
				{
					return table.inode[i];
				}
				else
				{
					free(inode.filename);
				}
			}
		}
		
	}while(table.indexNext != 0);
	return -1;
}
/* Renvoie l'offset du dernier object dans le systeme de fihier
 * 
 * fs	: Acces au systeme de fichier
 * type : type d'object a chercher : inode, data, adressetable, indextable,all
 * adresse_parent : permet de l'appelant quelle est le pere de l'object qu'il recupere data->inode  table_adresse->inode
 */
off_t search_last_object_in_fs(FileSystem_t * fs,enum type type,off_t *adresse_parent)
{
	IndexTable_t table = *(fs->index_table);
	inode_t varinode;
	adresse_table_t table_adresse;
	off_t addresse_table;
	int firsttour = 1;
	int i;
	off_t last = 0;
	do
	{
		if(firsttour)
		{
			firsttour = 0;
		}
		else
		{
			table = getindextable(fs->fd,table.indexNext);
		}
		if(type == indextable && table.indexNext > last)
		{
		    last = table.indexNext;
		}
		if(type == data || type == all || type == adressetable || type == inode)
		{
		  
		    for(i = 0; i < NB_ENTRY_INDEX_TABLE; i++)
		    {
		      
			    if(table.inode[i] != 0)
			    {
				    if((type == inode || type == all) && table.inode[i] > last)
				    {
					if(adresse_parent != NULL)
					{
					  *adresse_parent = table.inode[i];
					}
					last = table.inode[i];
				    }
				    
				    if(type == data || type == all || type == adressetable)
				    {
					varinode = read_inode_on_fs(fs->fd,table.inode[i]);
					int u;
					
					if(type == data || type == all)
					{
					  for(u = 0; u < NB_ENTRY_DATA_INODE ; u++)
					  {
					      if(varinode.data[u] > last)
					      {
						if(adresse_parent != NULL)
						{
						  *adresse_parent = table.inode[i];
						}
						last = varinode.data[u];
					      }
					  }
					}
					addresse_table =  varinode.table_adresse;
					free(varinode.filename);
					while(addresse_table != 0)
					{
					  if((type == adressetable || type == all) && addresse_table > last)
					  {
					    if(adresse_parent != NULL)
					    {
					      *adresse_parent = table.inode[i];
					    }
					      last = addresse_table;
					  }
					  table_adresse = read_adresse_table_on_fs(fs->fd,addresse_table);
					  if(type == data || type == all)
					  {
					    for(u = 0; u < NB_ENTRY_DATA_INODE ; u++)
					    {
					      if(table_adresse.data[u] > last)
					      {
						if(adresse_parent != NULL)
						{
						  *adresse_parent = table.inode[i];
						}
						last = table_adresse.data[u];
					      }
					    }
					  }
					  addresse_table = table_adresse.addresse_table_next;
					}
				    }
			    }
		    }
		}
		
	}while(table.indexNext != 0);
	if(last == 0)
	{
	  return -1;
	}
	return last;
}


