#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "FileSystem.h"

/* Ecrit l'inode sur le systeme de fichier
* 
* fd	: descripteur de fichier du systeme de fichier
* offset : adresse ou ecrire l'inode
* inode : inode a ecrire
*/
void write_inode_on_fs(int fd,off_t offset,inode_t * inode)
{
	
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	
	unsigned int lenght_char = (strlen(inode->filename)+1)*sizeof(char);
	
	
	if(write(fd, &lenght_char, sizeof(unsigned int)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	// puis on ecrit le nom de fichier
	if(write(fd, inode->filename, lenght_char) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	// puis on ecrit l'inode
	if(write(fd, inode, sizeof(inode_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
		exit(EXIT_FAILURE);
	}
	
}


/* Renvoie et lit l'inode sur le systeme de fichier
* 
* fd	: descripteur de fichier du systeme de fichier
* offset : adresse ou lire l'inode
*/
inode_t read_inode_on_fs(int fd,off_t offset)
{
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	inode_t inode;
	unsigned int taille_name;
	char *filename;
	// on lis la taille du nom de fichier
	if(read(fd, &taille_name, sizeof(unsigned int)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	
	filename = (char *)malloc(sizeof(char)*taille_name);
	
	// on lis le nom de fichier
	if(read(fd, filename, sizeof(char)*taille_name) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	// on lit la structure 
	if(read(fd, &inode, sizeof(inode_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	inode.filename = filename;
	return inode;
}
/* Affiche l'inode passé en parametre
* 
* inode	: Inode a afficher
*/
void print_inode(inode_t * inode)
{
	printf("Nom de fichier : %s \n taille fichier : %lld \n",inode->filename,(long long int)inode->size);
	int i;
	for(i = 0 ; i < NB_ENTRY_DATA_INODE; i++)
	{
		printf("Data %d : %lld \n",i,(long long int)inode->data[i]);
	}
	printf("Adresse table : %lld\n",(long long int)inode->table_adresse);
}


/* Renvoie la taille d'un inode sur le disque
* 
* fd	: descripteur de fichier du systeme de fichier
* offset : adresse ou se trouve l'inode
*/
off_t sizeofinode(int fd,off_t offset)
{
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	unsigned int taille_name;
	
	// on lis la taille du nom de fichier
	if(read(fd, &taille_name, sizeof(unsigned int)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	return (sizeof(unsigned int) + (taille_name * sizeof(char)) + sizeof(inode_t));
}

/* change l'enplacement d'un block de donner pour un inode donné
* 
* fd	: descripteur de fichier du systeme de fichier
* inode_off : adresse ou se trouve l'inode
* ancien_offset : ancien offset du block de data
* new_offset : nouvelle emplacement du block de data
*/
void editoffsetofdata(int fd,off_t inode_off,off_t ancien_offset,off_t new_offset)
{
	inode_t inode = read_inode_on_fs(fd,inode_off);
	adresse_table_t table;
	int u;
	off_t addresse_table =  inode.table_adresse;
	
	for(u = 0; u < NB_ENTRY_DATA_INODE ; u++)
	{
		if(inode.data[u] == ancien_offset)
		{
			inode.data[u] = new_offset;
			write_inode_on_fs(fd,inode_off,&inode);
			return;
		}
	}
	
	
	
	while(addresse_table != 0)
	{
		
		table = read_adresse_table_on_fs(fd,addresse_table);
		
		for(u = 0; u < NB_ENTRY_DATA_INODE ; u++)
		{
			if(table.data[u] == ancien_offset)
			{
				table.data[u] = new_offset;
				write_adresse_table_on_fs(fd,addresse_table,&table);
				return;
			}
		}
		
		addresse_table = table.addresse_table_next;
	}

}


/* change l'enplacement d'une table d'adresse pour un inode donné
* 
* fd	: descripteur de fichier du systeme de fichier
* inode_off : adresse ou se trouve l'inode
* ancien_offset : ancien offset de la table d'adresse
* new_offset : nouvelle emplacement de la table d'adresse
*/
void editoffsetofadressetable(int fd,off_t inode_off,off_t ancien_offset,off_t new_offset)
{
	inode_t inode = read_inode_on_fs(fd,inode_off);
	adresse_table_t table;
	off_t next_addresse_table =  inode.table_adresse;
	
	off_t last_addresse_table = 0;
	
	int i = 0;
	while(next_addresse_table != 0)
	{
		
		table = read_adresse_table_on_fs(fd,next_addresse_table);
		
		if(next_addresse_table == ancien_offset && i == 0)
		{
			inode.table_adresse = new_offset;
			write_inode_on_fs(fd,inode_off,&inode);
			return;
		}
		else if (next_addresse_table == ancien_offset)
		{
			table = read_adresse_table_on_fs(fd,last_addresse_table);
			table.addresse_table_next = new_offset;
			write_adresse_table_on_fs(fd,last_addresse_table,&table);
			return;
		}
		i++;
		last_addresse_table = next_addresse_table;
		next_addresse_table = table.addresse_table_next;
	}

}
/* libere l'inode
* 
* inode	: adresse de l'inode a liberer
*/
void freeInode(inode_t * inode)
{
	free(inode->filename);
	free(inode);
}

