#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "FileSystem.h"

/* Ecrit la table d'adresse a l'offset spécifier
* 
* fs	: Permet d'acceder au system de fichier
* offset	: Endroit ou ecrire la table
* table_adresse : Table d'adresse a écrire
*/
void write_adresse_table_on_fs(int fd,off_t offset,adresse_table_t * table_adresse)
{
	
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	
	// puis on ecrit la table d'adresse
	if(write(fd, table_adresse, sizeof(adresse_table_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de l’ecriture : ‘%s’.\n",strerror(errno));
		exit(EXIT_FAILURE);
	}
	
}
/* Lit et renvoie la table d'adresse de l'offset spécifier
* 
* fs	: Permet d'acceder au system de fichier
* offset	: Endroit ou lire la table
*/
adresse_table_t read_adresse_table_on_fs(int fd,off_t offset)
{
	// on place le curseur au bon endroit
	if(lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de le lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	adresse_table_t table_adresse;
	
	// on lit la structure 
	if(read(fd, &table_adresse, sizeof(adresse_table_t)) == -1)
	{
		fprintf(stderr, "L’erreur suivante a été rencontrée lors de la lecture du fichier : ‘%s’.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	return table_adresse;
}
/* Affiche une table d'adresse 
* 
* table_adresse	: Table d'adresse a afficher
*/
void print_adresse_table(adresse_table_t * table_adresse)
{
	int i;
	for(i = 0 ; i < NB_ENTRY_DATA_INODE; i++)
	{
		printf("Data %d : %lld \n",i,(long long int)table_adresse->data[i]);
	}
	printf("Prochaine table adresse : %lld\n",(long long int)table_adresse->addresse_table_next);
}

